# Waiting for Heads Example from Chapter 0 Section 8

from frplib.expectations import E
from frplib.kinds        import conditional_kind, constant, weighted_as
from frplib.quantity     import as_quantity
from frplib.statistics   import __
from frplib.symbolic     import symbol
from frplib.utils        import iterate
from frplib.vec_tuples   import as_scalar


def wait_for_heads(remaining_flips, q=symbol('q')):
    """Returns the Kind of the total number of flips to get a head.

    `remaining_flips` is the kind of the additional number of flips
        required after seeing a tails

    `q` is the probability of getting a tails. [Default: symbol('q')]

    """
    # Make sure q is a symbol or high-precision number
    q = as_quantity(q)

    # The kind of the current flip
    flip = weighted_as(0, 1, weights=[q, 1 - q])

    # the kind of the total number of flips given the current flip
    flips_given_current = conditional_kind({  # type: ignore
        0: remaining_flips ^ (__ + 1),
        1: constant(1)
    })

    total_flips = flips_given_current // flip
    return total_flips

def wait_more_than(n, q=symbol('q')):
    "Returns probability of waiting more than n flips for heads; q is the weight on tails."
    wait = iterate(wait_for_heads, n + 1, constant(1), q=q)
    probability = E(wait ^ (__ > n))
    return as_scalar( probability )

def wait_exactly(n, q=symbol('q')):
    "Returns probability of waiting more than n flips for heads; q is the weight on tails."
    wait = iterate(wait_for_heads, n + 2, constant(1), q=q)
    probability = E(wait ^ (__ == n))
    return as_scalar(probability)



# Recursive Rover Example from Chapter 0 Section 8

from frplib.kinds        import conditional_kind, constant, uniform
from frplib.statistics   import __


def time_to_base(t):
    """Returns the conditional kind of time to base.
    Here, t is the *kind* of the remaining time *after the step*.

    """
    base = conditional_kind({1: constant(3),   # type: ignore
                             2: t ^ (__ + 5),
                             3: t ^ (__ + 7)})
    channel = uniform(1, 2, 3)

    return base // channel



# Six of One, Equilateral of the Other Example in Chapter 0 Sec 8

__all__ = ['vertices', 'is_equilateral', 'vertex_dists',
           'side_lengths', 'heron', 'T_kind', 'A_kind',]

from frplib.kinds        import clean, without_replacement
from frplib.numeric      import numeric_abs, numeric_sqrt
from frplib.statistics   import statistic, Diff, Sqrt
from frplib.utils        import irange

vertices = without_replacement(3, irange(1, 6))
is_equilateral = vertices ^ (Diff == (2, 2))

vertex_dists = [0, 1, numeric_sqrt(3), 2, numeric_sqrt(3), 1]

@statistic
def side_lengths(triangle):
    return [vertex_dists[int(numeric_abs(triangle[i] - triangle[(i - 1) % 3]))] for i in range(3)]

@statistic
def heron(a, b, c):
    s = (a + b + c) / 2
    return Sqrt(s * (s - a) * (s - b) * (s - c))

T_kind = is_equilateral
A_kind = clean(vertices ^ side_lengths ^ heron)



from __future__ import annotations

import pytest

from frplib.expectations import E
from frplib.kinds        import weighted_as
from frplib.symbolic     import simplify, symbol

def test_symbolic_E():
    p = symbol('p')
    k0 = weighted_as(-1, 0, 1, weights=[1, p, p**2])
    assert E(k0).raw == simplify((p**2 - 1) / (1 + p + p**2))  # tests fix of Bug 20



from __future__ import annotations

import math
import pytest

from frplib.exceptions import DomainDimensionError, InputError
from frplib.statistics import (Statistic, Condition, MonoidalStatistic,
                               is_statistic, statistic, condition, scalar_statistic,
                               tuple_safe, infinity, ANY_TUPLE,
                               fork, chain, compose,
                               Id, Scalar, __, Proj, _x_,
                               Sum, Count, Max, Min, Mean, Abs,
                               Sqrt, Floor, Ceil,
                               Exp, Log, Log2, Log10,
                               Sin, Cos, Tan, ATan2, Sinh, Cosh, Tanh,
                               Diff, Diffs, Permute,
                               Constantly, Fork, ForEach, IfThenElse,
                               And, Or, Not, Xor, top, bottom,
                               )
from frplib.quantity   import as_quantity, qvec
from frplib.symbolic   import symbol
from frplib.vec_tuples import vec_tuple


def test_builtin_statistics():
    "Builtin statistics and combinators."

    assert (2 * __)(1) == vec_tuple(2)
    assert (2 * __)((1,)) == vec_tuple(2)
    assert (2 * __)((1, 2, 3)) == vec_tuple(2, 4, 6)

    assert (2 ** __)(1, 2, 3) == vec_tuple(2, 4, 8)
    assert (__ ** 2)(1, 2, 3) == vec_tuple(1, 4, 9)
    assert (__ % 2 == 0)(4) == vec_tuple(1)
    assert ForEach(__ % 2)(1, 2, 3, 4) == vec_tuple(1, 0, 1, 0)
    assert ForEach(__ % 2 == 0)(1, 2, 3, 4) == vec_tuple(0, 1, 0, 1)

    assert IfThenElse(Proj[1](__) == 2, 2 * __, __ + 10)(11, 12, 13) == vec_tuple(21, 22, 23)

    assert Fork(__, Constantly(2), __ ** 2)(1) == vec_tuple(1, 2, 1)
    assert Fork(__, Constantly(2), __ ** 2)((1,)) == vec_tuple(1, 2, 1)
    assert Fork(__, Constantly(2), __ ** 2)((1, 2, 3)) == vec_tuple(1, 2, 3, 2, 1, 4, 9)

    assert Cos(1)[0] == pytest.approx(as_quantity(math.cos(1)))
    assert Sin(1)[0] == pytest.approx(as_quantity(math.sin(1)))

    assert Scalar(Sin ** 2 + Cos ** 2) == pytest.approx(1)

    assert Sum((1, 2, 3, 4, 5)) == vec_tuple(15)
    assert Count((1, 2, 3, 4, 5)) == vec_tuple(5)
    assert Max((1, 2, 3, 4, 5)) == vec_tuple(5)
    assert Min((1, 2, 3, 4, 5)) == vec_tuple(1)

    assert Diff((1, 2, 3, 4, 5)) == vec_tuple(1, 1, 1, 1)
    assert Diffs(2)((1, 2, 3, 4, 5)) == vec_tuple(0, 0, 0)

    a = symbol('a')
    assert str((1 + 2 * __ + 3 * __ ** 2)(a)) == '<1 + 2 a + 3 a^2>'

    assert Proj[1](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(10)
    assert Proj[2](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(20)
    assert Proj[3](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(30)
    assert Proj[-1](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(80)
    assert Proj[3:6](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(30, 40, 50)
    assert Proj[3:-2](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(30, 40, 50, 60)
    assert Proj[-3:](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(60, 70, 80)
    assert Proj[:](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(10, 20, 30, 40, 50, 60, 70, 80)
    assert Proj[1, 2, 4, 8](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(10, 20, 40, 80)
    assert Proj[-7, -5, -3, -1](10, 20, 30, 40, 50, 60, 70, 80) == vec_tuple(20, 40, 60, 80)

    assert And(Proj[1] % 2 == 0, Proj[2] > 0)(-12, 21) == vec_tuple(1)
    assert And(Proj[1] % 2 == 0, Proj[2] > 0)(-11, 21) == vec_tuple(0)
    assert And(Proj[1] % 2 == 0, Proj[2] > 0)(-11, -21) == vec_tuple(0)
    assert And(Proj[1] % 2 == 0, Proj[2] > 0)(0, -21) == vec_tuple(0)
    assert And(Proj[1] % 2 == 0, Proj[2] > 0)(0, 0) == vec_tuple(0)
    assert And(Proj[1] % 2 == 0, Proj[2] > 0)(2, 2) == vec_tuple(1)

    assert Or(Proj[1] % 2 == 0, Proj[2] > 0)(-12, 21) == vec_tuple(1)
    assert Or(Proj[1] % 2 == 0, Proj[2] > 0)(-11, 21) == vec_tuple(1)
    assert Or(Proj[1] % 2 == 0, Proj[2] > 0)(-11, -21) == vec_tuple(0)
    assert Or(Proj[1] % 2 == 0, Proj[2] > 0)(0, -21) == vec_tuple(1)
    assert Or(Proj[1] % 2 == 0, Proj[2] > 0)(0, 0) == vec_tuple(1)
    assert Or(Proj[1] % 2 == 0, Proj[2] > 0)(1, 1) == vec_tuple(1)

    assert Not(Proj[1] % 2 == 0)(2) == vec_tuple(0)
    assert Not(Proj[1] % 2 == 0)(3) == vec_tuple(1)
    assert Not(Proj[1] % 2 == 0)(5) == vec_tuple(1)
    assert Not(Proj[1] % 2 == 0)(8) == vec_tuple(0)
    assert Not(Proj[1] % 2 == 0)(2, 7) == vec_tuple(0)
    assert Not(Proj[1] % 2 == 0)(3, 9) == vec_tuple(1)
    assert Not(Proj[1] % 2 == 0)(5, 9, 10) == vec_tuple(1)
    assert Not(Proj[1] % 2 == 0)(8, 3, 2, 1) == vec_tuple(0)

    assert Permute(1, 4, 2, 3)(10, 20, 30, 40, 50, 60, 70) == vec_tuple(10, 40, 20, 30, 50, 60, 70)

    assert Floor(-4.2) == vec_tuple(-5)
    assert Ceil(-4.2) == vec_tuple(-4)
    assert Floor(4.2) == vec_tuple(4)
    assert Ceil(4.2) == vec_tuple(5)
    assert Floor(0) == vec_tuple(0)
    assert Ceil(0) == vec_tuple(0)

    assert Abs(-4.2) == qvec(4.2)
    assert Abs('4.2') == Abs(4.2)

    assert Sqrt(4) == vec_tuple(2)
    assert Sqrt(1.44) == vec_tuple(as_quantity(1.2))

def test_tuple_safe():
    def sc_fn(x):
        return as_quantity(x) + 1   # type: ignore

    s1 = tuple_safe(sc_fn, arities=1, strict=False)
    s1s = tuple_safe(sc_fn, arities=1, strict=True)

    assert s1(4) == vec_tuple(5)
    assert s1((4,)) == vec_tuple(5)
    assert s1(-1) == vec_tuple(0)
    assert s1('1/2') == vec_tuple(1.5)
    assert s1(17, 10) == vec_tuple(18)
    assert s1((17, 10)) == vec_tuple(18)
    assert s1((-101, 1, 2, 3, 4, 5)) == vec_tuple(-100)

    with pytest.raises(DomainDimensionError):
        s1s(17, 10)

    with pytest.raises(DomainDimensionError):
        s1s((17, 10))

    with pytest.raises(DomainDimensionError):
        s1s()

    with pytest.raises(DomainDimensionError):
        s1s((1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12))

    def bad_1(x, y, z):
        return x + y + z

    with pytest.raises(InputError):
        tuple_safe(bad_1, arities=(1, 3), strict=False)

    with pytest.raises(InputError):
        tuple_safe(bad_1, arities=4, strict=False)

    with pytest.raises(InputError):
        tuple_safe(bad_1, arities=(0, infinity), strict=False)

    def v_fn(a, b, c):
        return (a, b, c, 0)

    v1 = tuple_safe(v_fn, strict=False)
    v1s = tuple_safe(v_fn, strict=True)

    assert v1(1, 2, 3) == vec_tuple(1, 2, 3, 0)
    assert v1((1, 2, 3)) == vec_tuple(1, 2, 3, 0)
    assert v1((1, 2, 3, 4, 5, 6, 7, 8)) == vec_tuple(1, 2, 3, 0)

    with pytest.raises(DomainDimensionError):
        v1s((1, 2, 3, 4))

    with pytest.raises(DomainDimensionError):
        v1(1, 2)

    with pytest.raises(DomainDimensionError):
        v1s(1, 2)

    with pytest.raises(DomainDimensionError):
        v1((4,))

    with pytest.raises(DomainDimensionError):
        v1()

    def vt_fn(a):
        u, v, w = a
        return (u, v, w, 0)

    v2 = tuple_safe(vt_fn, arities=3, strict=False)
    v2s = tuple_safe(vt_fn, arities=3, strict=True)

    assert v2(1, 2, 3) == vec_tuple(1, 2, 3, 0)
    assert v2((1, 2, 3)) == vec_tuple(1, 2, 3, 0)
    assert v2((1, 2, 3, 4, 5, 6, 7, 8)) == vec_tuple(1, 2, 3, 0)

    with pytest.raises(DomainDimensionError):
        v2s((1, 2, 3, 4))

    with pytest.raises(DomainDimensionError):
        v2((1, 2,))

    with pytest.raises(DomainDimensionError):
        v2(1, 2,)

    with pytest.raises(DomainDimensionError):
        v2()

    def vm_fn(a):
        return sum(a[2:])

    v3 = tuple_safe(vm_fn, arities=(3, infinity), strict=False)
    assert v3(1, 2, 3, 4) == vec_tuple(7)
    assert v3(tuple(range(11))) == vec_tuple(54)
    assert v3(10, 90, 80) == vec_tuple(80)

    with pytest.raises(DomainDimensionError):
        v3((1, 2))
    with pytest.raises(DomainDimensionError):
        v3(4)
    with pytest.raises(DomainDimensionError):
        v3()



#
# Labyrinth Example as described in the text
#

__all__ = [
    'labyrinth', 'steps', 'from_latest', 'start', 'moves',
    'n_moves', 'theseus_latest', 'nth', 'after_move_n',
    'ever_visited',
]

from collections.abc   import Generator, Iterator
from itertools         import islice
from typing            import TypeVar

from frplib.kinds      import ConditionalKind, conditional_kind, constant, Kind, uniform
from frplib.statistics import statistic
from frplib.vec_tuples import as_numeric_vec


# The Labyrinth

labyrinth: dict[int, list[int]] = {
    0: [1, 3],
    1: [0, 5],
    2: [4, 5],
    3: [0, 12],
    4: [2, 5],
    5: [1, 2, 4],
    6: [5, 7],
    7: [6, 8],
    8: [7, 9],
    9: [8, 10],
    10: [9, 33],
    11: [14, 32],
    12: [3, 17],
    13: [14],
    14: [11, 13],
    15: [16],
    16: [15, 18, 32],
    17: [12, 21, 32],
    18: [16, 19, 32],
    19: [18, 20],
    20: [19],
    21: [17, 22],
    22: [21, 23],
    23: [22, 24],
    24: [23, 26],
    25: [26],
    26: [24, 25, 27],
    27: [26, 28],
    28: [27, 29, 30],
    29: [28, 30],
    30: [28, 29, 31],
    31: [30],
    32: [11, 16, 17, 18],
    33: [33],  # Once we get to the end, we stay there
}


# Individual steps

steps = conditional_kind({
    juncture: uniform(neighbors)                    # type: ignore
    for juncture, neighbors in labyrinth.items()
})

assert isinstance(steps, ConditionalKind)

def from_latest(steps: ConditionalKind):
    def mixture(path):
        latest = path[-1]
        return steps[latest]
    return conditional_kind(mixture)


# Initial state mixer and state update targets

start = constant(0)
moves = from_latest(steps)


# Multiple Moves

def n_moves(start, moves, n):
    """Makes n moves from start and returns the kind up to that point.

    Warning: This kind grows large with n, so keep n to small values.

    See after_move_n and theseus_latest for a more efficient way to
    look at later moves.

    """
    current = start
    for _ in range(n):
        current = current >> moves
    return current

def theseus_latest(initial: Kind, moves: ConditionalKind) -> Generator:
    """Returns a generator of Theseus's latest move.

    Try, for instance:

      nth(theseus_latest(start, moves), 20)

    which gives the kind of Theseus's position after 20 moves.

    """
    current = initial
    while True:
        current = (current >> moves)[2]
        yield current

T = TypeVar('T')

def nth(seq: Iterator[T], n: int) -> T:
    "Extracts the nth item from an iterator."
    return next(islice(seq, n, n + 1))

def after_move_n(n: int, start: Kind, moves: ConditionalKind) -> Kind:
    "The kind of Theseus's state after n moves from start."
    return nth(theseus_latest(start, moves), n)

def ever_visited(n: int):
    """Returns a statistic that extracts the current room and visit counts from Theseus's path.

    Parameter n is the number of rooms, and it is assumed the rooms
    are labeled 0..n-1.

    """
    # convert to current room and binary array 1..n
    @statistic
    def visits(path: tuple[int, ...]) -> tuple[int, ...]:
        rooms = [0] * (n + 1)
        rooms[0] = path[-1]  # current room
        for room in path:
            rooms[1 + room] = 1
        return as_numeric_vec(rooms)
    return visits



# Buckets and Balls Example Chapter 0 Section 8

from typing             import cast
from frplib.kinds       import Kind, QuantityType, conditional_kind, bayes, either

bucket = either(0, 1)
green_given_bucket = conditional_kind(cast(dict[QuantityType, Kind], {
    0: either(0, 1, 9),
    1: either(0, 1, 4)
}))

which_bucket_g = bayes(observed_y=1, x=bucket, y_given_x=green_given_bucket)
which_bucket_n = bayes(observed_y=0, x=bucket, y_given_x=green_given_bucket)



# Markov Decryption Example from Section 6

__all__ = ['make_cipher', 'markov_decrypt', 'cipher1', 'clear1']

import math

from typing            import Callable, Union

from frplib.exceptions import InputError
from frplib.frps       import frp
from frplib.kinds      import ordered_samples, weighted_as
from frplib.utils      import clone

_CHARS = [
    ' ', 'A', 'B', 'C', 'D', 'E',
    'F', 'G', 'H', 'I', 'J',
    'K', 'L', 'M', 'N', 'O',
    'P', 'Q', 'R', 'S', 'T',
    'U', 'V', 'W', 'X', 'Y', 'Z',
]
_N_CHARS = len(_CHARS)
_END_MARK = _CHARS[0]  # Spaces, bos, and eos are treated as equivalent

# CITATION: The table _BIGRAM_LLIKE is derived from data produced
# by Peter Norvig, see http://....
# ATTN

_BIGRAM_LLIKE: dict[tuple[str, str], float] = {
    # ATTN:MISSING  # Process data tables
}

def make_cipher(substitution: Union[str, list[str]]) -> tuple[Callable[[str], str], Callable[[str], str]]:
    """Creates encryption and decryption functions for a specified substitution.

    Parameter `substitution` should be either a string or an array of single characters
    that contains a permutation of a space and the 26 capital English letters.

    Returns a tuple of functions str -> str: (encrypt, decrypt).

    """
    if isinstance(substitution, str):
        substitution = [c for c in substitution]

    if not all(c1 == c2 for c1, c2 in zip(_CHARS, sorted(substitution))):
        raise InputError('make_cipher requires a permutation of space and 26 capital letters')

    _enc = { c1: c2 for c1, c2 in zip(_CHARS, substitution) }
    _dec = { c2: c1 for c1, c2 in zip(_CHARS, substitution) }

    def encrypt(clear_text: str) -> str:
        return ''.join([_enc[c] for c in clear_text])

    def decrypt(cipher_text: str) -> str:
        return ''.join([_dec[c] for c in cipher_text])

    return (encrypt, decrypt)

def _log_like(text):
    ell = _BIGRAM_LLIKE[(_END_MARK, text[0])]

    for ind in range(_N_CHARS - 1):
        ell += _BIGRAM_LLIKE[(text[ind], text[ind + 1])]

    ell += _BIGRAM_LLIKE[(text[_N_CHARS - 1], _END_MARK)]

    return ell

def _occurs(event):
    return event.value[0] == 1

# ATTN: maybe allow n_best=1 and keep that many best states and scores
# when ==1 just return a single best, else a list of best in order

def markov_decrypt(cipher, iter=1000):
    state = _CHARS[:]
    best_score = _log_like([c for c in cipher])
    best_state = state[:]

    pair = frp(ordered_samples(2, list(range(_N_CHARS))))

    score = best_score
    for _ in range(iter):
        a, b = clone(pair).value
        candidate = state[:]
        candidate[a], candidate[b] = candidate[b], candidate[a]
        decrypt = { c2: c1 for c1, c2 in zip(_CHARS, candidate) }

        cand_score = _log_like([decrypt[c] for c in cipher])
        p = min(1, math.exp(cand_score - score))

        if p >= 1 or _occurs(frp(weighted_as(0, 1, weights=[1 - p, p]))):
            score = cand_score
            state = candidate

            if score > best_score:
                best_score = score
                best_state = state[:]

    decrypt = { c2: c1 for c1, c2 in zip(_CHARS, best_state) }
    decrypted = ''.join([decrypt[c] for c in cipher])

    return (decrypted, best_state, best_score)


# Simple Examples

cipher1 = 'QXYVE WMAVDOCBJVLCKVP TGYFVCHYOVQXYVRUSNVFCI'
clear1 = 'THE QUICK BROWN FOX JUMPED OVER THE LAZY DOG'



# Roulette Example Chapter 0, Section 2

__all__ = ['roulette']

from frplib.exceptions   import IndexingError
from frplib.frps         import frp
from frplib.kinds        import uniform
from frplib.statistics   import statistic

ROULETTE_SPIN = uniform(-1, 0, ..., 36)

RED_SQUARES = set([1, 3, 5, 7, 9, 12, 14, 16, 18,
                   19, 21, 23, 25, 27, 30, 32, 34, 36])

#
# Plays
#

# Even-Money Plays

@statistic(dim=1, codim=1)
def _roulette_even(pocket):
    "Play on all even pockets."
    if pocket % 2 == 0 and pocket >= 1 and pocket <= 36:
        return 1
    return -1

@statistic(dim=1, codim=1)
def _roulette_odd(pocket):
    "Play on all odd pockets."
    if pocket % 2 == 1 and pocket >= 1 and pocket <= 36:
        return 1
    return -1

@statistic(dim=1, codim=1)
def _roulette_red(pocket):
    "Play on all red pockets."
    if pocket in RED_SQUARES and pocket >= 1 and pocket <= 36:
        return 1
    return -1

@statistic(dim=1, codim=1)
def _roulette_black(pocket):
    "Play on all black pockets."
    if pocket not in RED_SQUARES and pocket >= 1 and pocket <= 36:
        return 1
    return -1

@statistic(dim=1, codim=1)
def _roulette_first18(pocket):
    "Play on first 18 consecutive pockets 1..18."
    if pocket >= 1 and pocket <= 18:
        return 1
    return -1

@statistic(dim=1, codim=1)
def _roulette_second18(pocket):
    "Play on second 18 consecutive pockets 19..36."
    if pocket >= 19 and pocket <= 36:
        return 1
    return -1

# 2-to-1 Plays

def _roulette_dozen(which):
    "Dozen play on twelve consecutive pockets in 1..36, specified by 1, 2, or 3, first, second, third ...."
    if which in [1, 2, 3]:
        which_dozen = which - 1
    elif isinstance(which, str):
        doz = which.lower()
        if doz in ['1', 'first', '1st']:
            which_dozen = 0
        elif doz in ['2', 'second', '2nd']:
            which_dozen = 1
        elif doz in ['3', 'third', '3rd']:
            which_dozen = 2
        else:
            raise IndexingError(f'Invalid Dozen play specifier: {which}. Try 1, 2, or 3 or first, second, or third.')
    else:
        raise IndexingError(f'Invalid Dozen play specifier {which}. Try 1, 2, or 3 or first, second, or third.')

    @statistic(dim=1, codim=1)
    def dozen_play(pocket):
        if which_dozen * 12 < pocket <= (which_dozen + 1) * 12:
            return 2
        return -1

    return dozen_play

def _roulette_column(which):
    "Column play on twelve pockets in one `column`, specified by 1, 2, or 3, first, second, third ...."
    if which == 3:
        which_column = 0
    elif which == 1 or which == 2:
        which_column = which
    elif isinstance(which, str):
        col = which.lower()
        if col in ['1', 'first', '1st']:
            which_column = 1
        elif col in ['2', 'second', '2nd']:
            which_column = 2
        elif col in ['3', 'third', '3rd']:
            which_column = 0
        else:
            raise IndexingError(f'Invalid Column play specifier: {which}. Try 1, 2, or 3 or first, second, or third.')
    else:
        raise IndexingError(f'Invalid Column play specifier {which}. Try 1, 2, or 3 or first, second, or third.')

    @statistic(dim=1, codim=1)
    def column_play(pocket):
        if 1 <= pocket <= 36 and pocket % 3 == which_column:
            return 2
        return -1

    return column_play

# Line Plays

def _roulette_six_line(first_row):
    "Six Line play on six pockets in two adjaced `rows`, specified by any pocket in smallest row."
    if not isinstance(first_row, int) or first_row < 1 or first_row > 36:
        raise IndexingError(f'Invalid pocket {first_row} to specify Six Line play, should be in 1..36.')

    @statistic(dim=1, codim=1)
    def six_line(pocket):
        start = 3 * ((first_row - 1) // 3)
        if start < pocket <= start + 6:
            return 5
        return -1

    return six_line

@statistic(dim=1, codim=1)
def _roulette_top_line(pocket):
    ""
    if pocket <= 3:
        return 6
    return -1

# Other Plays

def _roulette_corner(smallest):
    "Corner play specified by smallest square among four sharing a corner."
    if not isinstance(smallest, int) or smallest < 1 or smallest > 36:
        raise IndexingError(f'Invalid pocket {smallest} to specify Corner play. '
                            f'It should be the smallest square in 1..36 among four that share a corner.')

    winners = set([smallest, smallest + 1, smallest + 3, smallest + 4])

    @statistic(dim=1, codim=1)
    def corner(pocket):
        if pocket in winners:
            return 8
        return -1

    return corner

def _roulette_street(first_row):
    "Street play on three pockets in one `row` specified by any pocket in the row."
    if not isinstance(first_row, int) or first_row < 1 or first_row > 36:
        raise IndexingError(f'Invalid pocket {first_row} to specify Six Line play, should be in 1..36.')

    @statistic(dim=1, codim=1)
    def street(pocket):
        start = 3 * ((first_row - 1) // 3)
        if start < pocket <= start + 3:
            return 11
        return -1

    return street

def _roulette_split(first, second):
    "Split play on two adjacent pockets in -1, 0, 1..36, the first smaller than the second."
    if first < second and (second - first == 1 or second - first == 3
                           or (first == -1 and second in [0, 2, 3])
                           or (first == 0 and second in [1, 2])):

        @statistic(dim=1, codim=1)
        def split(pocket):
            if pocket == first or pocket == second:
                return 17
            return -1

        return split

    raise IndexingError(f'Invalid pair to specify a Split play {(first, second)}. '
                        f'they need to be adjacent with first < second.')

def _roulette_straight(wins):
    "Straight play on the specified pocket in -1, 0, 1..36."
    if not isinstance(wins, int) or wins < -1 or wins > 36:
        raise IndexingError(f'Invalid pocket {wins} to specify a straight play, should be in -1, 0, 1..36.')

    @statistic(dim=1, codim=1)
    def straight(pocket):
        if pocket == wins:
            return 35
        return -1

    return straight


#
# Entry Point
#

def roulette(n=1):
    """An interface to FRPs and statistics representing Roulette spins and plays.

    When called as a function, returns an FRP representing n spins (n=1 default).

    roulette.plays is a list of available standard plays that are available
    as statistics or statistic factories. For instance, roulette.even
    is the Even play and roulette.straight(20) is the Straight play on
    pocket 20.

    """
    return frp(ROULETTE_SPIN) ** n

setattr(roulette, 'plays',
        [
            'even', 'odd', 'red', 'black', 'first18', 'second18',
            'dozen', 'column', 'six_line', 'top_line',
            'corner', 'street', 'split', 'straight',
        ])

setattr(roulette, 'kind', ROULETTE_SPIN)

# Even-Money Plays

setattr(roulette, 'even',     _roulette_even)
setattr(roulette, 'odd',      _roulette_odd)
setattr(roulette, 'red',      _roulette_red)
setattr(roulette, 'black',    _roulette_black)
setattr(roulette, 'first18',  _roulette_first18)
setattr(roulette, 'second18', _roulette_second18)

# 2-to-1 Plays

setattr(roulette, 'dozen',  _roulette_dozen)
setattr(roulette, 'column', _roulette_column)

# Line Plays

setattr(roulette, 'six_line', _roulette_six_line)
setattr(roulette, 'top_line', _roulette_top_line)

# Other Plays

setattr(roulette, 'corner',   _roulette_corner)
setattr(roulette, 'street',   _roulette_street)
setattr(roulette, 'split',    _roulette_split)
setattr(roulette, 'straight', _roulette_straight)



from __future__ import annotations

from collections.abc   import Iterable
from functools         import wraps

from frplib.exceptions import ComplexExpectationWarning
from frplib.frps       import FRP, ConditionalFRP
from frplib.kinds      import Kind, ConditionalKind
from frplib.output     import in_panel
from frplib.protocols  import SupportsExpectation, SupportsApproxExpectation, SupportsForcedExpectation
from frplib.quantity   import show_qtuple
from frplib.statistics import Statistic
from frplib.vec_tuples import VecTuple, as_vec_tuple


class Expectation(VecTuple):
    def __init__(self, contents: Iterable):
        self.label = ''

    def __str__(self) -> str:
        return show_qtuple(self)

    def __frplib_repr__(self):
        return in_panel(str(self), title=self.label or None)

    # ATTN: needed?
    @property
    def raw(self):
        "Expectation as a raw scalar or vector tuple."
        if len(self) == 1:
            return self[0]
        return VecTuple(self)

def E(x, force_kind=False, allow_approx=True, tolerance=0.01):
    """Computes and returns the expectation of a given object.

    If `x` is an FRP or kind, its expectation is computed directly,
    unless doing so seems computationally inadvisable. In this case,
    the expectation is forced if `forced_kind` is True and otherwise
    is approximated, with specified `tolerance`, if `allow_approx`
    is True.

    In this case, returns a quantity wrapping the expectation that
    allows convenient display at the repl; the actual value is in
    the .this property of the returned object.

    If `x` is a ConditionalKind or ConditionalFRP, then returns
    a *function* from domain values in the conditional to
    expectations.

    """
    if isinstance(x, (ConditionalKind, ConditionalFRP)):
        f = x.expectation()

        @wraps(f)
        def c_expectation(*xs):
            return Expectation(as_vec_tuple(f(*xs)))
        if getattr(f, 'dim') is not None:
            label = f'dimension {getattr(f, "dim")} values'
        else:
            label = 'values'
        setattr(c_expectation, '__frplib_repr__',
                lambda: f'A conditional expectation as a function of {label}.')
        return c_expectation

    if isinstance(x, SupportsExpectation):
        label = ''
        try:
            expect = x.expectation()
        except ComplexExpectationWarning as e:
            if force_kind and isinstance(x, SupportsForcedExpectation):
                expect = x.forced_expectation()
            elif isinstance(x, SupportsApproxExpectation):
                expect = x.approximate_expectation(tolerance)
                label = (f'Computing approximation (tolerance {tolerance}) '
                         f'as exact calculation may be costly')  # ': {str(e)}\n'
            else:
                raise e
        expect = Expectation(as_vec_tuple(expect))
        if label:
            expect.label = label
        return expect
    return None

def D_(X: FRP | Kind):
    """The distribution operator for an FRP or kind.

    When passed an FRP or kind, this returns a function
    that maps any compatible statistic to the expectation
    of the transformed FRP or kind.

    """
    def probe(psi: Statistic):
        # ATTN: Check compatibility here
        return E(psi(X))
    return probe


#
# Info tags
#

setattr(E, '__info__', 'actions')
setattr(D_, '__info__', 'actions')



# Doubled Heads and Other Patterns Example from Chapter 0 Section 8

from frplib.kinds        import conditional_kind, constant, weighted_as
from frplib.quantity     import as_quantity
from frplib.statistics   import __
from frplib.symbolic     import symbol


def wait_for_2heads(remaining_flips, q=symbol('q')):
    """Returns the kind of the total number of flips to get consecutive heads.

    `remaining_flips` is the kind of the additional number of flips
        required after seeing a tails or heads-tails.

    `q` is the probability of getting a tails. [Default: symbol('q')]

    """
    # Make sure q is a symbol or high-precision number
    q = as_quantity(q)

    # The kind of the current prefix flips
    prefix = weighted_as(0, 2, 3, weights=[q, q * (1 - q), (1 - q) * (1 - q)])

    # the kind of the total number of flips given the current flip
    flips_given_current = conditional_kind({  # type: ignore
        0: remaining_flips ^ (__ + 1),
        2: remaining_flips ^ (__ + 2),
        3: constant(2)
    })

    total_flips = flips_given_current // prefix
    return total_flips



# Mouse Escape Example from Chapter 0 Section 8

__all__ = ['initial_state', 'move', 'escaped',
           'initial_state_alt', 'move_alt', 'mouse_outcome',
           'escaped_alt']

from frplib.expectations import E
from frplib.frps         import evolve
from frplib.kinds        import conditional_kind, constant, either
from frplib.statistics   import __, Proj


# First Approach

initial_state = constant(0)
move = conditional_kind({  # type: ignore
    0: either(0, 1, 2),
    1: either(0, 2, 2),
    2: either(1, 3, 2),
    3: constant(3)
})

state = initial_state
for _ in range(16):
    state = move // state
escaped = E(state ^ (__ == 3))


# Alternative Approach

initial_state_alt = constant(0, 0)

@conditional_kind
def move_alt(attempts_and_step):
    n_attempts, step = attempts_and_step

    # If we are at the end, stay there
    if step == 3 or step == -1:
        return constant(attempts_and_step)

    n = n_attempts + 1

    # If the cat is here, last chance
    if n_attempts == 16:
        if step < 2:
            return constant(n_attempts, -1)
        else:
            return either((n, -1), (n, 3), 2)

    # From step 0, we either stay or move up
    if step == 0:
        return either((n, 0), (n, 1), 2)

    # Otherwise, we move up or down
    return either((n, step - 1), (n, step + 1), 2)

mouse_outcome = evolve(initial_state_alt, move_alt, 16)   # type: ignore
escaped_alt = E(mouse_outcome ^ (Proj[2] == 3))



# A Kid Named Florida Example Chapter 0 Section 8

from frplib.calculate   import substitution
from frplib.kinds       import clean, either, weighted_as
from frplib.statistics  import And, Or, Proj
from frplib.symbolic    import symbol

# Question (a)

at_least_one_girl = Or(Proj[1] == 1, Proj[2] == 1)
both_girls = And(Proj[1] == 1, Proj[2] == 1)
girl = either(0, 1)

outcome_no_names = girl * girl | at_least_one_girl


# Question (b)

a_rare_girl = Or(
    And(Proj[1] == 1, Proj[3] == 11),
    And(Proj[2] == 1, Proj[4] == 11)
)

p = symbol('p')
rare = weighted_as(10, 11, weights=[1 - p, p])
neighbors = girl * girl * rare * rare

outcome = neighbors | a_rare_girl
outcome_zero = clean(substitution(outcome, p=0))



# Doubled Cards Example, Chapter 0 Section 8

from frplib.frps         import frp
from frplib.kinds        import conditional_kind, uniform, ordered_samples
from frplib.statistics   import Or, Proj
from frplib.utils        import irange


first_card = uniform(1, 2, ..., 100)
all_cards = first_card.values
second_card = conditional_kind({
    first: uniform(all_cards - {first}) for first in all_cards
})

# The Kind of the drawn pair of cards
draw = first_card >> second_card
draw_alt = ordered_samples(2, irange(1, 100))

# Conditions that test for our desired outcomes
is_card_doubled = Proj[2] == 2 * Proj[1]
is_either_doubled = Or(is_card_doubled, Proj[1] == 2 * Proj[2])

# Relevant FRPs and Kinds
P = frp(draw)
D = is_card_doubled(P)
T = is_either_doubled(P)

D_kind = is_card_doubled(draw)     # same as kind(D)
T_kind = is_either_doubled(draw)   # same as kind(T)



from __future__ import annotations

from itertools import chain

from frplib.examples.roulette import roulette, RED_SQUARES

from frplib.kinds      import Kind, kind, weighted_as
from frplib.utils      import irange
from frplib.vec_tuples import vec_tuple

def test_roulette_basics():
    R = roulette()
    kind_R = kind(R)
    assert Kind.equal(roulette.kind, kind_R)
    assert Kind.equal(roulette.even(kind_R),
                      weighted_as(-1, 1, weights=['10/19', '9/19']))
    assert Kind.equal(roulette.straight(16)(kind_R),
                      weighted_as(-1, 35, weights=['37/38', '1/38']))

def test_roulette_plays():
    lost = vec_tuple(-1)
    won = vec_tuple(1)    # scale as needed

    assert all( roulette.even(p) == won for p in irange(2, 36, step=2) )
    assert all( roulette.even(p) == lost for p in irange(-1, 36, step=2) )
    assert roulette.even(0) == lost
    assert all( roulette.odd(p) == won for p in irange(1, 36, step=2) )
    assert all( roulette.odd(p) == lost for p in irange(2, 36, step=2) )
    assert roulette.odd(0) == lost
    assert roulette.odd(-1) == lost

    assert all( roulette.red(p) == won for p in RED_SQUARES )
    assert all( roulette.red(p) == lost for p in irange(-1, 36) if p not in RED_SQUARES )
    assert all( roulette.black(p) == won for p in irange(1, 36) if p not in RED_SQUARES )
    assert all( roulette.black(p) == lost for p in RED_SQUARES )
    assert roulette.black(0) == lost
    assert roulette.black(-1) == lost

    assert all( roulette.first18(p) == won for p in irange(1, 18) )
    assert all( roulette.second18(p) == won for p in irange(19, 36) )
    assert all( roulette.first18(p) == lost for p in irange(-1, 36) if p < 1 or p > 18)
    assert all( roulette.second18(p) == lost for p in irange(-1, 18) )

    assert all( roulette.dozen(1) == 2 * won for p in irange(1, 12) )
    assert all( roulette.dozen(2) == 2 * won for p in irange(13, 24) )
    assert all( roulette.dozen(3) == 2 * won for p in irange(25, 36) )
    assert all( roulette.dozen('first') == 2 * won for p in irange(1, 12) )
    assert all( roulette.dozen('second') == 2 * won for p in irange(13, 24) )
    assert all( roulette.dozen('third') == 2 * won for p in irange(25, 36) )
    assert all( roulette.dozen(1) == lost for p in chain([-1, 0], irange(13, 36)) )
    assert all( roulette.dozen(2) == lost for p in chain(irange(-1, 12), irange(25, 36)) )
    assert all( roulette.dozen(3) == lost for p in irange(-1, 24) )

    assert all( roulette.column(1) == 2 * won for p in irange(1, 36) if p % 3 == 1 )
    assert all( roulette.column(2) == 2 * won for p in irange(1, 36) if p % 3 == 2 )
    assert all( roulette.column(3) == 2 * won for p in irange(1, 36) if p % 3 == 0 )
    assert all( roulette.column(1) == lost    for p in irange(1, 36) if p % 3 != 1 )
    assert all( roulette.column(2) == lost    for p in irange(1, 36) if p % 3 != 2 )
    assert all( roulette.column(3) == lost    for p in irange(1, 36) if p % 3 != 0 )
    assert all( roulette.column(i)(0) == lost  for i in irange(1, 3) )
    assert all( roulette.column(i)(-1) == lost for i in irange(1, 3) )

    assert all( roulette.top_line == 5 * won for p in irange(-1, 3) )
    assert all( roulette.top_line == lost for p in irange(4, 36) )

    assert all( roulette.six_line(3 * u + k)(3 * u + p) == 5 * won
                for u in range(11) for k in irange(1, 3) for p in irange(1, 6) )
    assert all( roulette.six_line(p)(-1) == lost for p in irange(1, 36) )
    assert all( roulette.six_line(p)(0) == lost for p in irange(1, 36) )
    assert all( roulette.six_line(3 * u + k)(p) == lost
                for u in range(11) for k in irange(1, 3) for p in irange(1, 3 * u) )
    assert all( roulette.six_line(3 * u + k)(p) == lost
                for u in range(11) for k in irange(1, 3) for p in irange(3 * u + 7, 36) )

    assert all( roulette.street(3 * u + k)(3 * u + p) == 11 * won
                for u in range(11) for k in irange(1, 3) for p in irange(1, 3) )
    assert all( roulette.street(p)(-1) == lost for p in irange(1, 36) )
    assert all( roulette.street(p)(0) == lost for p in irange(1, 36) )
    assert all( roulette.street(3 * u + k)(p) == lost
                for u in range(11) for k in irange(1, 3) for p in irange(1, 3 * u) )
    assert all( roulette.street(3 * u + k)(p) == lost
                for u in range(11) for k in irange(1, 3) for p in irange(3 * u + 4, 36) )

    assert all( roulette.corner(3 * u + k)(x) == 8 * won for u in range(11) for k in [1, 2]
                for x in [3 * u + k, 3 * u + k + 1, 3 * u + k + 3, 3 * u + k + 4] )
    assert all( roulette.corner(3 * u + k)(x) == lost for u in range(11) for k in [1, 2]
                for x in irange(-1, 36, exclude={3 * u + k, 3 * u + k + 1, 3 * u + k + 3, 3 * u + k + 4}) )

    assert all( roulette.split(3 * u + k, p)(3 * u + k) == 17 * won
                for u in range(11) for k in irange(1, 2) for p in [3 * u + k + 1, 3 * u + k + 3] )
    assert all( roulette.split(3 * u + k, p)(p) == 17 * won
                for u in range(11) for k in irange(1, 2) for p in [3 * u + k + 1, 3 * u + k + 3] )
    assert all( roulette.split(3 * u, 3 * u + 1)(p) == 17 * won
                for u in irange(1, 11) for p in [3 * u, 3 * u + 1])
    assert all( roulette.split(u, u + 1)(p) == 17 * won
                for u in irange(34, 35) for p in [u, u + 1] )
    assert all( roulette.split(3 * u + k, p)(3 * u + k) == 17 * won
                for u in range(11) for k in irange(1, 2) for p in [3 * u + k + 1, 3 * u + k + 3] )
    assert all( roulette.split(3 * u + k, p)(x) == lost
                for u in range(11) for k in irange(1, 2) for p in [3 * u + k + 1, 3 * u + k + 3]
                for x in irange(-1, 36, exclude={3 * u + k, 3 * u + k + 1, 3 * u + k + 3}) )
    assert all( roulette.split(3 * u, 3 * u + 1)(x) == lost
                for u in irange(1, 11) for p in [3 * u, 3 * u + 1]
                for x in irange(-1, 36, exclude={3 * u, 3 * u + 1}) )
    assert all( roulette.split(u, u + 1)(p) == lost
                for u in irange(34, 35) for p in irange(-1, 36, exclude={u, u + 1}) )

    assert all( roulette.straight(p)(p) == 35 * won for p in irange(-1, 36) )
    assert all( roulette.straight(p)(q) == lost for p in irange(-1, 36) for q in irange(-1, 36) if p != q )

def test_text():
    R = roulette()
    s16_10 = 10 * roulette.straight(16)
    W16_10 = R ^ s16_10
    assert Kind.equal(kind(W16_10), weighted_as(-10, 350, weights=['37/38', '1/38']))

    comb = 10 * roulette.even + 5 * roulette.corner(25) + 20 * roulette.straight(4) + 50 * roulette.column(2)
    assert comb(26)[0] == 130

    k = weighted_as(-85, -65, -40, -20, 65, 85, 110, 130, 655,
                    weights=['13/38', '10/38', '1/38', '1/38', '5/38', '5/38', '1/38', '1/38', '1/38'])
    assert Kind.equal(kind(comb(R)), k)



# Disease Testing Redux Example from Chapter 0 Section 8

from frplib.kinds       import bayes, conditional_kind, weighted_as
from frplib.symbolic    import symbol


d = symbol('d')
n = symbol('n')
p = symbol('p')

has_disease = weighted_as(0, 1, weights=[1 - d, d])
test_by_status = conditional_kind({             # type: ignore
    0: weighted_as(0, 1, weights=[n, 1 - n]),
    1: weighted_as(0, 1, weights=[1 - p, p])
})

disease_given_positive = bayes(1, has_disease, test_by_status)
disease_given_negative = bayes(0, has_disease, test_by_status)



#
# Aces Example in Section 7
#

__all__ = ['between_aces', 'two_four_gaps', 'cut',]

from frplib.statistics import statistic, Id
from frplib.utils      import irange
from frplib.vec_tuples import as_vec_tuple


def _ace_gaps(deck, aces):
    "Computes tuple of gap sizes between specified `aces` in a deck."
    gaps = []
    deck_size = len(deck)
    start = 0
    for pos, card in enumerate(deck):
        if card in aces:
            gaps.append(pos - start)
            start = pos + 1
    gaps.append(deck_size - start)

    return as_vec_tuple(gaps)

@statistic
def between_aces(deck):
    "Returns a tuple of the number of cards between successive aces (including the ends of the deck)."
    return _ace_gaps(deck, {1, 14, 27, 40})   # These *are* the cards we're looking for

@statistic
def two_four_gaps(deck):
    "A simple analogue of between_aces for a deck of five cards and two 'aces'."
    return _ace_gaps(deck, {2, 4})

def cut(ace1, ace2, deck_size=52, aces={1, 14, 27, 40}):
    """A statistic factory that swaps the segments between two aces specified by order.

    Parameters
      ace1: int - ordinal in 1..len(aces)+1 of an ace (len(aces)+1 for end of deck)
      ace2: int - ordinal in 1..len(aces)+1 of another ace (len(aces)+1 for end of deck)
      deck_size: int [=52] - number of cards in the deck 1..deck_size
      aces: set[int] [={1,14,27,40}] - cards in deck corresponding to aces

    Returns a statistic that modifies a deck by swapping the cards
    strictly between ace1 and the previous ace (or beginning of deck)
    and strictly between ace2 and the previous ace (or beginning of deck).
    If ace1 or ace2 is len(aces)+1, the segment after the last ace is used.

    If ace1 and ace2 are not distinct, this is just the identity statistic.


    """
    aces = set(aces)  # ensure we have a set
    end_mark = len(aces) + 1

    if ace1 == ace2:
        return Id
    if ace1 > ace2:
        ace1, ace2 = ace2, ace1

    @statistic
    def do_cut(deck):
        swapped = [0] * deck_size
        ace_pos = [-1] + [pos for pos, card in enumerate(deck) if card in aces] + [deck_size]

        index = 0
        for ace in irange(1, end_mark):
            if ace == ace1:
                ace_cap = ace2
            elif ace == ace2:
                ace_cap = ace1
            else:
                ace_cap = ace

            # Move the segment before the designated ace
            n = ace_pos[ace_cap] - ace_pos[ace_cap - 1] - 1
            swapped[index:(index + n)] = deck[(ace_pos[ace_cap - 1] + 1):ace_pos[ace_cap]]
            index += n

            if ace < end_mark:  # Move the ace also
                swapped[index] = deck[ace_pos[ace]]
                index += 1

        return as_vec_tuple(swapped)

    return do_cut



from __future__ import annotations

import math

from collections.abc   import Iterable
from decimal           import Decimal
from fractions         import Fraction
from functools         import reduce
from operator          import add, mul, sub
from typing            import cast, Type, TypeVar, Union
from typing_extensions import Self, TypeGuard

from frplib.exceptions import OperationError, NumericConversionError
from frplib.numeric    import Numeric, NumericF, NumericD, NumericB, numeric_sqrt  # ATTN: Numeric+Symbolic+SupportsVec
from frplib.numeric    import as_numeric as scalar_as_numeric
from frplib.symbolic   import Symbolic, symbolic_sqrt

# SupportsVec  mixin can allow Symbolic and VecTuple automatically,   __plus__  __scalar_mul__
# SupportsNumeric protocol __numeric__ with numeric conversion.
# Can newtype Decimal and Fraction etc.; subclass with __slots__

#
# Types
#

# A VecTuple should contain entirely interoperable types
T = TypeVar('T', NumericF, NumericD, NumericB, Union[Numeric, Symbolic])


#
# Helpers
#

def extend(x, k, scalar_only=False):
    # Want T here but allow Quantity here when available
    if isinstance(x, (int, float, Fraction, Decimal, Symbolic)):
        return VecTuple([x] * k)
    if not scalar_only and isinstance(x, (tuple, list)) and len(x) == 1:
        return VecTuple([*x] * k)
    return x

def from_scalar(x):
    # Want T here but allow Quantity here when available
    if isinstance(x, (int, float, Fraction, Decimal, Symbolic)):
        return VecTuple([x])
    return x

def as_scalar(x) -> T | None:
    if isinstance(x, (int, float, Fraction, Decimal, Symbolic, bool)):
        return cast(T, x)
    elif isinstance(x, tuple) and len(x) == 1:
        return cast(T, x[0])
    return None

def as_scalar_strict(x) -> T:
    if isinstance(x, (int, float, Fraction, Decimal, Symbolic, str, bool)):
        return cast(T, x)
    elif isinstance(x, tuple) and len(x) == 1:
        return cast(T, x[0])
    raise NumericConversionError(f'The quantity {x} could not be converted to a numeric/symbolic scalar.')


#
# Numeric/Quantified VecTuples
#

class VecTuple(tuple[T, ...]):
    "A variant tuple type that supports addition and scalar multiplication like a vector."
    def __new__(cls, contents: Iterable[T]) -> 'VecTuple[T]':
        return super().__new__(cls, contents)     # type: ignore

    def __str__(self):
        return f'<{", ".join(map(str, self))}>'

    def __frplib_repr__(self):
        return self.__str__()

    def map(self, fn) -> 'VecTuple[T]':
        return self.__class__(map(fn, self))

    @property
    def dim(self):
        return len(self)

    @classmethod
    def show(cls, vtuple: 'VecTuple[T]', scalarize=True) -> str:
        if scalarize and len(vtuple) == 1:
            return str(vtuple[0])
        return str(vtuple)

    def __add__(self, other) -> 'VecTuple[T]':
        other = extend(other, len(self))  # Experimental: R style vector extension
        if not isinstance(other, tuple):
            return NotImplemented
        return VecTuple(map(add, self, other))

    def __radd__(self, other) -> 'VecTuple[T]':
        other = extend(other, len(self))  # Experimental: R style vector extension
        if not isinstance(other, tuple):
            return NotImplemented
        return VecTuple(map(add, other, self))

    def __sub__(self, other) -> 'VecTuple[T]':
        other = extend(other, len(self))  # Experimental: R style vector extension
        if not isinstance(other, tuple):
            return NotImplemented
        return VecTuple(map(sub, self, other))

    def __rsub__(self, other) -> 'VecTuple[T]':
        other = extend(other, len(self))  # Experimental: R style vector extension
        if not isinstance(other, tuple):
            return NotImplemented
        return VecTuple(map(sub, other, self))

    def __mul__(self, other) -> 'VecTuple[T]':
        z: T | None = as_scalar(other)   # type: ignore
        if z is not None:
            # Since other is Any in the superclass, the type system
            # cannot let us check that other has a matching type
            try:
                return VecTuple(map(lambda x: x * z, self))     # type: ignore
            except Exception:
                return NotImplemented
        return NotImplemented

    def __rmul__(self, other) -> 'VecTuple[T]':
        z: T | None = as_scalar(other)   # type: ignore
        if z is not None:
            # Since other is Any in the superclass, the type system
            # cannot let us check that other has a matching type
            try:
                return VecTuple(map(lambda x: z * x, self))     # type: ignore
            except Exception:
                return NotImplemented
        return NotImplemented

    def __truediv__(self, other) -> 'VecTuple[T]':
        z: T | None = as_scalar(other)   # type: ignore
        if z is not None:
            # Since other is Any in the superclass, the type system
            # cannot let us check that other has a matching type
            try:
                return VecTuple(map(lambda x: x / z, self))  # type: ignore
            except Exception:
                return NotImplemented
        return NotImplemented

    def __floordiv__(self, other) -> 'VecTuple[T]':
        z: T | None = as_scalar(other)   # type: ignore
        if z is not None:
            # Since other is Any in the superclass, the type system
            # cannot let us check that other has a matching type
            try:
                return VecTuple(map(lambda x: x // z, self))  # type: ignore
            except Exception:
                return NotImplemented
        return NotImplemented

    def __mod__(self, other) -> 'VecTuple[T]':
        z: T | None = as_scalar(other)   # type: ignore
        if z is not None:
            # Since other is Any in the superclass, the type system
            # cannot let us check that other has a matching type
            try:
                return VecTuple(map(lambda x: x % z, self))  # type: ignore
            except Exception:
                return NotImplemented
        return NotImplemented

    def __pow__(self, other) -> 'VecTuple[T]':
        z: T | None = as_scalar(other)   # type: ignore
        if z is not None:
            # Since other is Any in the superclass, the type system
            # cannot let us check that other has a matching type
            try:
                return VecTuple(map(lambda x: x ** z, self))  # type: ignore
            except Exception:
                return NotImplemented
        return NotImplemented

    def __rpow__(self, other) -> 'VecTuple[T]':
        z: T | None = as_scalar(other)   # type: ignore
        if z is not None:
            # Since other is Any in the superclass, the type system
            # cannot let us check that other has a matching type
            try:
                return VecTuple(map(lambda x: z ** x, self))  # type: ignore
            except Exception:
                return NotImplemented
        return NotImplemented

    def __matmul__(self, other) -> T:
        if not isinstance(other, Iterable):  # Allow vector like things
            return NotImplemented
        return reduce(add, map(mul, self, other), cast(T, 0))

    def __abs__(self):
        sq_norm = self @ self
        if isinstance(sq_norm, Symbolic):
            raise NotImplementedError('Symbolic function application not yet implemented')
        dot_prod = self @ self
        if isinstance(dot_prod, Symbolic):
            return symbolic_sqrt(dot_prod)
        elif isinstance(dot_prod, (int, Decimal)):
            return numeric_sqrt(dot_prod)
        return math.sqrt(dot_prod)

    def __getitem__(self, key):
        x = super().__getitem__(key)
        if isinstance(key, slice):
            return VecTuple(x)
        return x

    def __eq__(self, other):
        other = from_scalar(other)   # Allow scalar equality of VecTuples, no invariants changed
        try:
            return super().__eq__(other)
        except TypeError as e:
            raise OperationError(f'Could not test for == with {other}: {str(e)}')

    def __hash__(self):  # Need this because we play with __eq__
        return super().__hash__()  # Modified __eq__ does not change the invariant

    def __ne__(self, other):
        other = from_scalar(other)   # Allow scalar comparison of VecTuples
        try:
            return super().__ne__(other)
        except TypeError as e:
            raise OperationError(f'Could not test for != with {other}: {str(e)}')

    def __lt__(self, other):
        other = from_scalar(other)   # Allow scalar comparison of VecTuples
        try:
            return super().__lt__(other)
        except TypeError as e:
            raise OperationError(f'Could not test for != with {other}: {str(e)}')

    def __le__(self, other):
        other = from_scalar(other)   # Allow scalar comparison of VecTuples
        try:
            return super().__le__(other)
        except TypeError as e:
            raise OperationError(f'Could not test for != with {other}: {str(e)}')

    def __gt__(self, other):
        other = from_scalar(other)   # Allow scalar comparison of VecTuples
        try:
            return super().__gt__(other)
        except TypeError as e:
            raise OperationError(f'Could not test for != with {other}: {str(e)}')

    def __ge__(self, other):
        other = from_scalar(other)   # Allow scalar comparison of VecTuples
        try:
            return super().__ge__(other)
        except TypeError as e:
            raise OperationError(f'Could not test for != with {other}: {str(e)}')

    @classmethod
    def join(cls: Type[Self], values: Iterable[Self]) -> Self:
        combined = []
        for value in values:
            combined.extend(list(value))
        return cls(combined)


def vec_tuple(*a: T) -> VecTuple[T]:
    "Collects its arguments into a VecTuple"
    return VecTuple(a)

def as_vec_tuple(x: T | Iterable[T] = ()) -> VecTuple[T]:
    "Converts an iterable to (or wraps a single value in) a VecTuple"
    if isinstance(x, VecTuple):
        return x
    if isinstance(x, Iterable) and not isinstance(x, str):
        return VecTuple(x)
    return vec_tuple(x)

def as_numeric_vec(x):
    # ATTN: Consider using as_real here
    if isinstance(x, Iterable) and not isinstance(x, str):
        return VecTuple(map(scalar_as_numeric, x))
    else:
        return vec_tuple(scalar_as_numeric(x))

def is_vec_tuple(x) -> TypeGuard[VecTuple[T]]:
    "Is this a VecTuple?"
    return isinstance(x, VecTuple)



from __future__ import annotations

from decimal import Decimal

from frplib.kinds      import Kind, kind, weighted_as
from frplib.symbolic   import symbols
from frplib.statistics import IfThenElse, Norm, Proj
from frplib.utils      import irange, lmap
from frplib.vec_tuples import vec_tuple

from frplib.examples.monty_hall    import switch_win, dont_switch_win, outcome_by_strategy
from frplib.examples.circle_points import circle_points

def test_monty_hall():
    assert Kind.equal(dont_switch_win, weighted_as(0, 1, weights=['2/3', '1/3']))
    assert Kind.equal(switch_win, weighted_as(0, 1, weights=['1/3', '2/3']))

    a, b = symbols('a b')
    K0 = outcome_by_strategy(left=a, middle=b, right=1 - a - b)
    K1 = weighted_as([vec_tuple(i, j) for i in irange(3) for j in irange(3)],
                     weights=[a/3, (1 - a - b)/3, b/3, a/3, (1 - a - b)/3, b/3, a/3, (1 - a - b)/3, b/3])
    assert Kind.equal(K0, K1)

def test_circle_points():
    XY = circle_points()

    norm_vals = lmap(vec_tuple,
                     [Decimal('0'), Decimal('1'), Decimal('1.414213562373095048801688724'), Decimal('2'),
                      Decimal('2.236067977499789696409173669'), Decimal('2.828427124746190097603377448'),
                      Decimal('3'), Decimal('3.162277660168379331998893544'), Decimal('3.605551275463989293119221267'),
                      Decimal('4'), Decimal('4.123105625617660549821409856'),
                      Decimal('4.242640687119285146405066173'), Decimal('4.472135954999579392818347337'),
                      Decimal('5')])
    k = weighted_as(norm_vals,
                    weights=[0.012346, 0.049383, 0.049383, 0.049383, 0.098765, 0.049383, 0.049383,
                             0.098765, 0.098765, 0.049383, 0.098765, 0.049383, 0.098765, 0.14815])
    assert Kind.equal(kind(Norm(XY)), k, tolerance=1e-5)

    compareXY = IfThenElse(Proj[1] > Proj[2], 1, IfThenElse(Proj[1] < Proj[2], -1, 0))
    assert Kind.equal(kind(compareXY(XY)),
                      weighted_as(-1, 0, 1, weights=[0.45679, 0.086420, 0.45679]),
                      tolerance=1e-6)



# SPDX-FileCopyrightText: 2023-present Christopher R. Genovese <genovese@cmu.edu>
#
# SPDX-License-Identifier: MIT
import sys

if __name__ == "__main__":
    from frplib.cli import frp

    sys.exit(frp())



# Rig at Risk Example in Chapter 0 Section 8

from frplib.expectations import E
from frplib.kinds        import conditional_kind, fast_mixture_pow
from frplib.statistics   import Sum

def rig_at_risk(kind_N, kind_D):
    """Computes expectation of total wave damage by mixing kinds then computing E.
    Parameters `kind_N` and `kind_D` give arbitrary kinds for the number of waves
    and the damage per wave.

    """
    @conditional_kind(codim=1)
    def damage_given_n(n):
        return fast_mixture_pow(Sum, kind_D, n)

    return E(damage_given_n // kind_N)



# Elevator Stops Example in Chapter 0 Section 8

from frplib.kinds        import Kind, conditional_kind, constant, uniform, weighted_as
from frplib.statistics   import Statistic, statistic, Sum
from frplib.vec_tuples   import as_vec_tuple

def passengers_kind(p, mu):
    pvalues = list(range(p + 1))

    weights = [1] * (p + 1)
    w = 1
    for k in pvalues[1:]:
        weights[k] = w
        w *= mu / k

    return weighted_as(pvalues, weights=weights)

def visited_floors(top_floor: int) -> Statistic:
    "Returns a statistic converting a list of floor choices to a set."
    @statistic(name=f'visited_floors<{top_floor}>')
    def visited_set(value):
        "returns the set of unique components in a fixed range, as a bit string"
        bits = [0] * top_floor
        for x in value:       # values are floors in 1, 2, ..., top_floor
            bits[x - 1] = 1   # this floor's button has been pushed
        return as_vec_tuple(bits)
    return visited_set

def union_visited(top_floor: int) -> Statistic:
    "Returns a statistic that unions two `top-floor` sets as bit-strings."
    @statistic(name=f'union<{top_floor}>', monoidal=as_vec_tuple([0] * top_floor))
    def union(value):
        "unions two `top_floor`-length bit strings into one with a bitwise-or"
        return as_vec_tuple(value[i] | value[i + 10] for i in range(top_floor))
    return union

# Main analysis

def elevator_stops(max_floor=10, p=21, mu=5) -> Kind:
    as_set: Statistic = visited_floors(max_floor)  # Statistic: convert floors to visited sets
    union: Statistic = union_visited(max_floor)    # Statistic: Union of two sets as 10-bit strings

    passenger_floor = uniform(1, 2, ..., max_floor)
    choice = as_set(passenger_floor)  # Kind for each floor choice as set

    # Kind of visited floor set for each # passengers.
    floors: dict[int, Kind] = {0: constant([0] * max_floor), 1: choice, 2: union(choice * choice)}
    for i in range(2, p):
        floors[i + 1] = union(floors[i] * floors[1])
    visited = conditional_kind(floors)  # type: ignore

    number_passengers = passengers_kind(p, mu)
    number_visited_floors = Sum(visited) // number_passengers

    return number_visited_floors



from __future__ import annotations

import pytest

from frplib.frps       import frp, conditional_frp
from frplib.kinds      import Kind, kind, constant, either, uniform
from frplib.statistics import __, Proj
from frplib.utils      import dim, const

def test_empty_conditional():
    X = frp(uniform(1, 2, ..., 8))
    a = X.value
    Y = X | (__ < a)
    Z = X | (__ <= a)

    assert dim(Y) == 0
    assert Z.value == X.value

def test_frp_transform():
    X = frp(uniform(0, 1, ..., 7))
    assert Kind.equal(kind(X ^ (__ + 1)), uniform(1, 2, ..., 8))
    assert Kind.equal(kind(X ^ const(0)), constant(0))
    assert Kind.equal(kind(X * X ^ Proj[1]), kind(X))
    assert Kind.equal(kind(X * X ^ Proj[2]), kind(X))


#
# Tests of Conditional FRPs and related operations
#

def test_conditional_frps():
    u = conditional_frp({0: frp(either(0, 1)), 1: frp(uniform(1, 2, 3)), 2: frp(uniform(4, 5))})  # type: ignore
    v = frp(uniform(0, 1, 2))

    assert Kind.equal(kind(v >> u ^ Proj[2]), kind(u // v))  # tests fix of Bug 10



# Dirichlet Solutions Example from Section 6

__all__ = ['solve_dirichlet', 'solve_dirichlet_sparse', 'K_NSEW', 'lava_room']

from frplib.exceptions import InputError
from frplib.frps       import frp
from frplib.kinds      import conditional_kind, uniform, weighted_as
from frplib.utils      import clone, irange


@conditional_kind
def K_NSEW(tile):
    x, y = tile
    return uniform( (x - 1, y), (x, y - 1), (x, y + 1), (x + 1, y) )


def solve_dirichlet(cKind, *, fixed, fixed_values, alpha=0, beta=1, states=None):
    """Solves a Dirichlet problem determined by a conditional Kind and boundary constraints.

    Specifically, we want to solve for a function f on the domain of cKind
    that satisfies

       f(s) = fixed_values[i] when s in fixed[i] for some i, and
       f(s) = alpha + beta E(f(cKind(s))) otherwise.

    Parameters
      + cKind: ConditionalKind - determines Kind of transition from each state.
            Its domain is the set of possible states if explicitly available
            and the states parameter is not supplied.
      + fixed: list[set] - disjoint subsets of states on which f's value is known
      + fixed_values: list[float] - known values of f corresponding to fixed set
            in the same position. Must have the same length as fixed.
      + alpha: float [=0] - step cost parameter
      + beta: float [=1] - scaling parameter
      + states: None | Iterable - if supplied, the set of states that defines the domain
            of the function f.  If not supplied, must be obtainable explicitly
            from cKind.

    Returns a function of states (as tuples or multiple arguments)
    representing the solution f.

    """
    pass

def solve_dirichlet_sparse(cKind, *, fixed, fixed_values, alpha=0, beta=1, states=None):
    """Solves a Dirichlet problem determined by a conditional Kind and boundary constraints.

    Specifically, we want to solve for a function f on the domain of cKind
    that satisfies

       f(s) = fixed_values[i] when s in fixed[i] for some i, and
       f(s) = alpha + beta E(f(cKind(s))) otherwise.

    Parameters
      + cKind: ConditionalKind - determines Kind of transition from each state.
            Its domain is the set of possible states if explicitly available
            and the states parameter is not supplied.
      + fixed: list[set] - disjoint subsets of states on which f's value is known
      + fixed_values: list[float] - known values of f corresponding to fixed set
            in the same position. Must have the same length as fixed.
      + alpha: float [=0] - step cost parameter
      + beta: float [=1] - scaling parameter
      + states: None | Iterable - if supplied, the set of states that defines the domain
            of the function f.  If not supplied, must be obtainable explicitly
            from cKind.

    Returns a function of states (as tuples or multiple arguments)
    representing the solution f.

    """
    pass


# Example System in Text

class LavaRoom:
    'A room filled with lava and cool water arranged on a regular grid'

    def __init__(self):
        cells = [(x, y) for y in irange(-6, 6) for x in irange(-12, 12)]

        self.states = cells
        self.lava = {
            (x, y) for x, y in cells
            if ((8 <= x <= 12 and -6 <= y <= -5) or (x == 12 and -4 <= y <= -3) or
                (8 <= x <= 12 and y == 1) or (8 <= x < 9 and y == 0) or
                (11 <= x <= 12 and y == 0) or (x == -12 and -6 <= y <= 1) or
                (x == -11 and -6 <= y <= -1) or (x == -10 and y == -6) or
                (x == 6 and -8 <= y <= -7))
        }
        self.water = {
            (x, y) for x, y in cells
            if ((-8 <= x <= -5 and -6 <= y <= -3) or (6 <= x <= 7 and -6 <= y <= -5) or
                (-4 <= x <= 4 and y == 6) or (-1 <= x <= 1 and y == 5) or
                (x == 0 and y == 4))
        }
        self.fixed = [self.lava, self.water]
        # Originally free was a set, but a list is more useful # set(self.states) - self.lava - self.water

        self.free = [(100, 100)] * (len(cells) - len(self.lava) - len(self.water))
        self.state_index: dict[tuple[int, int], int] = {}
        self.free_index: dict[tuple[int, int], int] = {}
        j = 0
        for i, v in enumerate(cells):
            self.state_index[v] = i
            if v not in self.lava and v not in self.water:
                self.free[j] = v
                self.free_index[v] = j
                j += 1

    def _row(self, fcell, fixed_values=(0, 1)):
        "Returns row and RHS of the specified index (free cell index)."
        x, y = self.free[fcell]
        row: list[float] = [0.0] * len(self.free)
        rhs: float = 0

        row[fcell] = 1.0

        if abs(x) == 12 and abs(y) == 6:
            r = 0.5
        elif abs(x) == 12 or abs(y) == 6:
            r = 1.0 / 3.0
        else:
            r = 0.25

        for delta in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
            neighbor = (x + delta[0], y + delta[1])
            if abs(neighbor[0]) > 12 or abs(neighbor[1]) > 6:
                continue
            if neighbor in self.lava:
                rhs += r * fixed_values[0]
            elif neighbor in self.water:
                rhs += r * fixed_values[1]
            else:
                row[self.free_index[neighbor]] = -r

        return (row, rhs)

lava_room = LavaRoom()



# Numeric types that are interoperable but (ideally) readable and efficient
#
# These three categories can be used for probabilities or values.
# Prefer NumericF for probabilities, but this can get computationally costly,
# so considering using NumericD and post-converting to Fraction for output,
# conditionally on the denominator size. Get interpretable and accurate output
# without too much effort. Include these conversion functions here;
# See also VecTuple.py (formerly VecTupleSafe)

from __future__  import annotations

import math
import re

from abc               import abstractmethod
from collections.abc   import Iterable
from dataclasses       import dataclass
from decimal           import Decimal, ROUND_HALF_UP, ROUND_UP, ROUND_FLOOR, ROUND_CEILING
from enum              import Enum, auto
from fractions         import Fraction
from typing            import cast, Literal, Union
from typing_extensions import TypeAlias, TypeGuard

from frplib.env        import environment
from frplib.exceptions import EvaluationError


#
# Constants
#

DEFAULT_RATIONAL_DENOM_LIMIT = 1000000000   # Default Decimal -> Fraction conversion

class NumType(Enum):
    INTEGER = auto()
    RATIONAL = auto()
    REAL = auto()


#
# Numeric Quantities
#

class NumericQuantity:
    @abstractmethod
    def __float__(self) -> float:
        ...

    @abstractmethod
    def rational(self, limit_denominator=False) -> 'RationalQuantity':
        ...

    @abstractmethod
    def real(self) -> 'RealQuantity':
        ...

    def __lt__(self, other):
        if isinstance(other, NumericQuantity):
            return self.real() < other.real()
        return float(self) < other

    def __le__(self, other):
        if isinstance(other, NumericQuantity):
            return self.real() <= other.real()
        return float(self) <= other

    def __gt__(self, other):
        if isinstance(other, NumericQuantity):
            return self.real() > other.real()
        return float(self) > other

    def __ge__(self, other):
        if isinstance(other, NumericQuantity):
            return self.real() >= other.real()
        return float(self) >= other

@dataclass(frozen=True)
class IntegerQuantity(NumericQuantity):
    type: Literal[NumType.INTEGER] = NumType.INTEGER
    value: int = 0

    def __int__(self):
        return self.value

    def __index__(self):
        return self.value

    def __float__(self) -> float:
        return float(self.value)

    def rational(self, limit_denominator=False) -> 'RationalQuantity':
        return RationalQuantity(value=Fraction(self.value))

    def real(self) -> 'RealQuantity':
        return RealQuantity(value=Decimal(self.value))

    def __str__(self) -> str:
        return str(self.value)

@dataclass(frozen=True)
class RationalQuantity(NumericQuantity):
    type: Literal[NumType.RATIONAL] = NumType.RATIONAL
    value: Fraction = Fraction(0)

    def __float__(self) -> float:
        return float(self.value)

    def rational(self, limit_denominator=False) -> 'RationalQuantity':
        return self

    def real(self) -> 'RealQuantity':
        return RealQuantity(value=Decimal(self.value.numerator) / Decimal(self.value.denominator))

    def __str__(self) -> str:
        return str(self.value)

@dataclass(frozen=True)
class RealQuantity(NumericQuantity):
    type: Literal[NumType.REAL] = NumType.REAL
    value: Decimal = Decimal('0')

    def __float__(self) -> float:
        return float(self.value)

    def rational(self, limit_denominator=False) -> RationalQuantity:
        x = Fraction(self.value)
        if limit_denominator:
            x = x.limit_denominator(limit_denominator)
        return RationalQuantity(value=x)

    def real(self) -> 'RealQuantity':
        return self

    def __str__(self) -> str:
        return str(self.value)


#
# Numeric Types
#
# The NumericX types provide separate interoperable types for
# consistent arithmetic

NumericQ: TypeAlias = Union[IntegerQuantity, RationalQuantity, RealQuantity]
ScalarQ:  TypeAlias = Union[int, float, Fraction, Decimal, NumericQ, str]

NumericB: TypeAlias = Union[int, float]     # Binary floating point numbers
NumericD: TypeAlias = Union[int, Decimal]   # Decimal floating point numbers
NumericF: TypeAlias = Union[int, Fraction]  # Arbitrary Precision rational numbers

Numeric:  TypeAlias = NumericD  # Default underlying numeric representation

def is_scalar_q(x) -> TypeGuard[Union[int, float, Fraction, Decimal, NumericQ, str]]:
    return isinstance(x, (int, float, Fraction, Decimal, NumericQuantity, str, bool))  # bool auto cast to int

def is_numeric(x) -> TypeGuard[Union[int, Decimal]]:
    return isinstance(x, (int, Decimal))


#
# Numeric Conversion
#

REAL_ZERO = Decimal('0')
REAL_ONE = Decimal('1')
DECIMAL_DIG = 27  # current decimal precision used (digits)
NICE_DIGITS = 16  # used in nice_round; must be 0 <= _ <= DECIMAL_DIG with current Decimal setup

rat_denom = r'/(?:[1-9][0-9]{0,2}(?:_[0-9]{3})+|[1-9][0-9]*)'
decimal = r'\.[0-9]*'
sci_exp = r'[eE][-+]?(?:0|[1-9][0-9]*)'
opt_sign = r'-?'

integer_re = r'(?:0|[1-9][0-9]{0,2}(?:_[0-9]{3})+|[1-9][0-9]*)'  # _ separators allowed
numeric_re = rf'({opt_sign})({integer_re})(?:({rat_denom})|({decimal}(?:{sci_exp})?)|({sci_exp}))?'

def numeric_q_from_str(s: str) -> NumericQ:
    m = re.match(numeric_re, s.strip().replace('_', ''))
    if not m:
        m = re.match(r'(?i)-?inf(?:inity)?', s)
        if m:
            return RealQuantity(value=Decimal(s))  # +/- Infinity
        raise EvaluationError(f'Could not parse string as a numeric quantity: "{s}"')

    sign, integer, denom, dec_exp, exp = m.groups('')

    if not denom and not dec_exp and not exp:
        return IntegerQuantity(value=int(sign + integer))

    if denom:
        return RationalQuantity(value=Fraction(int(sign + integer), int(denom[1:])))

    if dec_exp:
        return RealQuantity(value=Decimal(sign + integer + dec_exp))

    return RealQuantity(value=Decimal(sign + integer + exp))

def numeric_q(
        x: ScalarQ = 0,
        exclude: Literal[NumType.RATIONAL] | Literal[NumType.REAL] | None = None,
        limit_denominator=DEFAULT_RATIONAL_DENOM_LIMIT
) -> NumericQ:
    if isinstance(x, str):
        x = numeric_q_from_str(x)    # Fall through

    if isinstance(x, IntegerQuantity):
        return x

    if isinstance(x, int):
        return IntegerQuantity(value=x)

    if isinstance(x, RationalQuantity):
        if exclude == NumType.RATIONAL:
            return x.real()
        return x

    if isinstance(x, RealQuantity):
        if exclude == NumType.REAL:
            return x.rational()
        return x

    if isinstance(x, Fraction):           # Check complexity to decide if real?
        qvalue = RationalQuantity(value=x)
        if exclude == NumType.RATIONAL:
            return qvalue.real()
        return qvalue

    if isinstance(x, float):
        val = nice_round(Decimal(x), NICE_DIGITS).normalize()
    else:
        val = x  # This is a Decimal

    rvalue = RealQuantity(value=val)
    if exclude == NumType.REAL:
        return rvalue.rational()
    return rvalue

def nice_round(d: Decimal, dig=NICE_DIGITS) -> Decimal:
    sign, digits, exp = d.as_tuple()
    dig = min(dig - 1, DECIMAL_DIG)  # ATTN: negative dig is ok but think about that case
    n = len(digits)
    if n <= dig + 1 or not isinstance(exp, int):  # exp can be n, N, or F
        return d
    true_exp = n - 1 + exp
    new_exp = true_exp - dig
    dtuple = (0, tuple([1] + ([0] * dig)), new_exp)
    return d.quantize(Decimal(dtuple), ROUND_HALF_UP)  # .normalize()

# Adjust these parameters for a default numerical quantification policy
def as_numeric(x: ScalarQ = RealQuantity()) -> Numeric:
    "A specialized version of `numeric` that defines a system-wide quantification policy."
    return cast(
        Union[IntegerQuantity, RealQuantity],
        numeric_q(x, exclude=NumType.RATIONAL, limit_denominator=1000000000)
    ).value

def as_rational(x: ScalarQ = RationalQuantity(), limit_denominator=False) -> Fraction:
    return numeric_q(x).rational(limit_denominator).value

def as_real(x: ScalarQ = RealQuantity()) -> Decimal:
    return numeric_q(x).real().value

def as_nice_numeric(x: ScalarQ = RealQuantity(), digits=NICE_DIGITS) -> Numeric:
    xn = as_numeric(x)
    if isinstance(xn, int):
        return xn
    return nice_round(xn, digits).normalize()


#
# Specialized Calculations
#

def numeric_sqrt(x: ScalarQ) -> Numeric:
    return as_real(x).sqrt()

def numeric_exp(x: ScalarQ) -> Numeric:
    return as_real(x).exp()

def numeric_ln(x: ScalarQ) -> Numeric:
    return as_real(x).ln()

def numeric_log10(x: ScalarQ) -> Numeric:
    return as_real(x).log10()

def numeric_log2(x: ScalarQ) -> Numeric:
    c = as_real(2).ln()
    return as_real(x).ln() / c

def numeric_abs(x: ScalarQ) -> Numeric:
    if isinstance(x, int):
        return abs(x)
    return as_real(x).copy_abs()

def numeric_floor(x: ScalarQ) -> Numeric:
    if isinstance(x, int):
        return x
    return as_real(x).quantize(REAL_ONE, ROUND_FLOOR)

def numeric_ceil(x: ScalarQ) -> Numeric:
    if isinstance(x, int):
        return x
    return as_real(x).quantize(REAL_ONE, ROUND_CEILING)


#
# Friendly Decimals
#
# This type behaves well at the repl, but to keep things light weight
# it only is useful to wrap the final values of a computation because
# Decimal operations return new Decimals.
#

class RealValue(Decimal):
    def __frplib_repr__(self) -> str:
        return str(self)

    def __repr__(self) -> str:
        if environment.is_interactive:
            return str(self)
        return super().__repr__()

#
# Numeric output (needs improvement)
#

# ATTN: Move these to environment as settable options, the first two
DEC_DENOMINATOR_LIMIT: int = 10**9
MAX_DENOMINATOR_EXC = 50
EXCLUDE_DENOMINATOR = {10, 20, 25, 50, 100, 125, 250, 500, 1000}
ROUND_MASK = Decimal('1.000000000')

def nroundx(x: Numeric, mask=ROUND_MASK, rounding=ROUND_HALF_UP) -> Decimal:
    return Decimal(x).quantize(mask, rounding)

def nround(x: Numeric, mask=ROUND_MASK, rounding=ROUND_HALF_UP) -> Decimal:
    return nroundx(x, mask, rounding).normalize()

def as_frac(x: Numeric, denom_limit=DEC_DENOMINATOR_LIMIT) -> Fraction:
    return Fraction(x).limit_denominator(denom_limit)

def denom_rules(denominator, max_denom=MAX_DENOMINATOR_EXC, exclude=EXCLUDE_DENOMINATOR) -> bool:
    return denominator < max_denom and denominator not in exclude

def show_prob(p: Numeric) -> str:
    "Provide a human readable view of a fractional number in [0,1]."
    if isinstance(p, int):
        return str(p)

    pf = as_frac(p)
    if denom_rules(pf.denominator):
        return str(pf)

    return str(nround(p))

def show_numeric(
        x: Numeric,
        max_denom=MAX_DENOMINATOR_EXC,
        exclude_denoms=EXCLUDE_DENOMINATOR,
        rounding_mask=ROUND_MASK,
        rounding=ROUND_HALF_UP
) -> str:
    if isinstance(x, int):
        return str(x)

    frac = as_frac(x)
    if denom_rules(frac.denominator, max_denom, exclude_denoms):
        return str(frac)

    return str(nround(x, rounding_mask, rounding))

def show_nice_numeric(
        x: Numeric,
        max_denom=MAX_DENOMINATOR_EXC,
        exclude_denoms=EXCLUDE_DENOMINATOR,
        digits=NICE_DIGITS
) -> str:
    if isinstance(x, int):
        return str(x)

    frac = as_frac(x)
    if denom_rules(frac.denominator, max_denom, exclude_denoms):
        return str(frac)

    return str(nice_round(x, digits).normalize())

def show_values(
        xs: Iterable[Numeric],
        max_denom=MAX_DENOMINATOR_EXC,
        exclude_denoms=EXCLUDE_DENOMINATOR,
        rounding_mask=ROUND_MASK,
        rounding=ROUND_HALF_UP,
        digit_shift=5
) -> list[str]:
    "Find a pleasing common representation for a list of numeric quantities."
    # Do the values all have a simple common denominator?
    # If so, show as rationals with common denominator
    # ATTN: If there are *two* shared denominators that follow the rules, that would be useful
    xs = list(xs)
    ratl_xs = [as_rational(x, limit_denominator=DEC_DENOMINATOR_LIMIT) for x in xs]
    common_denom = math.lcm(*[r.denominator for r in ratl_xs])
    if common_denom == 1:
        return list(map(str, ratl_xs))
    elif denom_rules(common_denom, max_denom, exclude_denoms):
        return [f'{int(x.numerator * common_denom / x.denominator)}/{common_denom}' for x in ratl_xs]

    # Otherwise, show as real numbers, rounding all to a reasonable common scale
    real_xs = [as_real(x) for x in xs]
    expts = [x.copy_abs() for x in real_xs if x != REAL_ZERO]
    size = (sum(expts) / len(expts)).log10() if len(expts) > 0 else digit_shift  # type: ignore
    digits = digit_shift - int(nround(size, REAL_ONE, ROUND_UP))
    mask = Decimal('1.' + ('0' * digits)) if digits > 0 else REAL_ONE

    return [str(nice_round(x, 5)) for x in real_xs]
    # return [str(nround(x, mask)) for x in real_xs]

def show_tuples(
        tups: Iterable[tuple],
        max_denom=MAX_DENOMINATOR_EXC,
        exclude_denoms=EXCLUDE_DENOMINATOR,
        rounding_mask=ROUND_MASK,
        rounding=ROUND_HALF_UP,
        scalarize=True
) -> list[str]:
    "Convert a list of tuples to strings with angle-bracket syntax, with a shared representation for each component."
    # if dim == 1:
    #     return show_values([tup[0] for tup in tups], max_denom, exclude_denoms, rounding_mask, rounding)

    # Transpose, Format, and Transpose back
    outT = []
    for out in zip(*tups):  # , strict=True
        outT.append(show_values(out, max_denom, exclude_denoms, rounding_mask, rounding))
    dim = len(outT)
    if scalarize and dim == 1:
        return [components[0] for components in zip(*outT)]  # , strict=True
    return [f'<{", ".join(components)}>' for components in zip(*outT)]  # , strict=True

# Use VecTuple.show when value types are normalized here
def show_tuple(
        tup: tuple,
        max_denom=MAX_DENOMINATOR_EXC,
        exclude_denoms=EXCLUDE_DENOMINATOR,
        rounding_mask=ROUND_MASK,
        rounding=ROUND_HALF_UP,
        scalarize=True
) -> str:
    "Show a tuple with angle bracket syntax, but drop brackets for scalars."
    if scalarize and len(tup) == 1:
        return show_numeric(tup[0], max_denom, exclude_denoms, rounding_mask, rounding)
    components = [show_numeric(x, max_denom, exclude_denoms, rounding_mask, rounding)
                  for x in tup]
    return f'<{", ".join(components)}>'


# Older

skip_denoms = {5, 10, 50, 100, 500, 1000, 5000, 10000}
def show_num(number: int | float | Decimal | Fraction,
             real_only=False, denom_limit=1000000, dec_round='1.00000000') -> str:
    if isinstance(number, float):
        if not real_only:
            rational = Fraction.from_float(number).limit_denominator(denom_limit)
            if rational.denominator not in skip_denoms and abs(number - float(rational)) < 1.0e-7:
                return str(rational)
        return str(Decimal(number).quantize(Decimal(dec_round), rounding=ROUND_HALF_UP).normalize())
    return str(number)

# Use VecTuple.show when value types are normalized here
def old_show_tuple(tup: tuple, real_only=False, denom_limit=1000000, dec_round='1.000000') -> str:
    "Show a tuple with angle bracket syntax, but give values only for scalars."
    if len(tup) == 1:
        return show_num(tup[0], real_only, denom_limit, dec_round)
    return f'<{", ".join([show_num(x, real_only, denom_limit, dec_round) for x in tup])}>'
